import pytest

from dennis.tools import (
    VariableTokenizer,
    PythonFormat,
    PythonBraceFormat,
    parse_dennis_note,
)


def test_empty_tokenizer():
    vartok = VariableTokenizer([])
    assert vartok.contains("python-format") is False
    assert vartok.tokenize("a b c d e") == ["a b c d e"]
    assert vartok.extract_tokens("a b c d e") == set()
    assert vartok.is_token("{0}") is False
    assert vartok.extract_variable_name("{0}") is None


@pytest.mark.parametrize(
    "text,expected",
    [
        ("Hello %s", ["Hello ", "%s"]),
        ("Hello %(username)s", ["Hello ", "%(username)s"]),
        ("Hello %(user)s%(name)s", ["Hello ", "%(user)s", "%(name)s"]),
        ("Hello {username}", ["Hello ", "{username}"]),
        ("Hello {user}{name}", ["Hello ", "{user}", "{name}"]),
        ("Products and Services", ["Products and Services"]),
    ],
)
def test_python_tokenizing(text, expected):
    vartok = VariableTokenizer(["python-format", "python-brace-format"])
    assert vartok.tokenize(text) == expected


class TestPythonBraceFormat:
    @pytest.mark.parametrize(
        "text,expected",
        [
            ("Hello", ["Hello"]),
            ("{foo}", ["{foo}"]),
            ("Hello {foo}", ["Hello ", "{foo}"]),
            ("{foo} Hello", ["{foo}", " Hello"]),
            ("{foo:%Y-%m-%d}", ["{foo:%Y-%m-%d}"]),
            ("{foo:%Y-%m-%d %H:%M}", ["{foo:%Y-%m-%d %H:%M}"]),
        ],
    )
    def test_parse(self, text, expected):
        vartok = VariableTokenizer(["python-brace-format"])
        assert vartok.tokenize(text) == expected

    @pytest.mark.parametrize(
        "text,expected",
        [
            ("{}", ""),
            ("{0}", "0"),
            ("{abc}", "abc"),
            ("{abc.def}", "abc.def"),
            ("{abc[0]}", "abc[0]"),
            ("{abc!s}", "abc"),  # conversion
            ("{abc: >16}", "abc"),  # format_spec
        ],
    )
    def test_variable_name(self, text, expected):
        v = PythonBraceFormat()
        assert v.extract_variable_name(text) == expected


@pytest.mark.parametrize(
    "text,expected", [("%s", ""), ("%d", ""), ("%.2f", ""), ("%(foo)s", "foo")]
)
def test_pythonformat(text, expected):
    v = PythonFormat()
    assert v.extract_variable_name(text) == expected


@pytest.mark.parametrize(
    "text,expected",
    [
        ("", []),
        ("Foo", []),
        ("Foo bar", []),
        ("dennis-ignore", []),
        ("dennis-ignore: *", "*"),
        ("dennis-ignore: E101", ["E101"]),
        ("dennis-ignore: E101, E102", ["E101"]),
        ("dennis-ignore: E101,E102", ["E101", "E102"]),
        ("localizers ignore this: dennis-ignore: E101,E102", ["E101", "E102"]),
    ],
)
def test_parse_dennis_note(text, expected):
    assert parse_dennis_note(text) == expected



POFILE_HEADER = """\
#, fuzzy
msgid ""
msgstr ""
"Project-Id-Version: foo\\n"
"POT-Creation-Date: 2013-06-05 14:16-0700\\n"
"PO-Revision-Date: 2010-04-26 18:00-0700\\n"
"Last-Translator: Automatically generated\\n"
"Language-Team: English\\n"
"Language: \\n"
"MIME-Version: 1.0\\n"
"Content-Type: text/plain; charset=UTF-8\\n"
"Content-Transfer-Encoding: 8bit\\n"
"X-Generator: Translate Toolkit 1.6.0\\n"
"""


def build_po_string(data):
    return POFILE_HEADER + "\n" + data


def nix_header(pot_file):
    """Nix the POT file header since it changes and we don't care"""
    return pot_file[pot_file.find("\n\n") + 2 :]



import os
import sys
import traceback
from functools import wraps
from textwrap import dedent

import click

from dennis import __version__
from dennis.linter import Linter
from dennis.linter import get_lint_rules as get_linter_rules
from dennis.templatelinter import TemplateLinter
from dennis.templatelinter import get_lint_rules as get_template_linter_rules
from dennis.tools import (
    get_available_formats,
    parse_pofile,
    withlines,
)
from dennis.translator import get_available_pipeline_parts, InvalidPipeline, Translator

USAGE = "%prog [options] [command] [command-options]"
VERSION = "dennis " + __version__


def utf8_args(fun):
    @wraps(fun)
    def _utf8_args(*args):
        args = [part.encode("utf-8") for part in args]
        return fun(*args)

    return _utf8_args


def err(s):
    """Prints a single-line string to stderr."""
    click.secho(f"Error: {s}", fg="red", bold=True, err=True)


def format_formats():
    formats = sorted(get_available_formats().items())
    lines = ["Available Variable Formats:", "", "\b"]
    lines.extend(
        ["{name:19}  {desc}".format(name=name, desc=cls.desc) for name, cls in formats]
    )

    text = "\n".join(lines) + "\n"
    return text


def format_pipeline_parts():
    parts = sorted(get_available_pipeline_parts().items())
    lines = ["Available Pipeline Parts:", "", "\b"]
    lines.extend(
        ["{name:13}  {desc}".format(name=name, desc=cls.desc) for name, cls in parts]
    )
    text = "\n".join(lines) + "\n"
    return text


def format_lint_rules():
    from dennis.linter import get_lint_rules

    rules = sorted(get_lint_rules().items())
    lines = ["Available Lint Rules:", "", "\b"]
    lines.extend(
        [
            "{num:6} {name}: {desc}".format(num=num, name=cls.name, desc=cls.desc)
            for num, cls in rules
        ]
    )
    text = "\n".join(lines) + "\n"
    return text


def format_lint_template_rules():
    from dennis.templatelinter import get_lint_rules

    rules = sorted(get_lint_rules().items())
    lines = ["Available Template Lint Rules:", "", "\b"]
    lines.extend(
        [
            "{num:6} {name}: {desc}".format(num=num, name=cls.name, desc=cls.desc)
            for num, cls in rules
        ]
    )
    text = "\n".join(lines) + "\n"
    return text


def epilog(docstring):
    """Fixes the docstring so it displays properly

    I have problems with docstrings and indentation showing up right
    in click. This fixes that. Plus it allows me to add dynamically
    generated bits to the docstring.

    """

    def _epilog(fun):
        fun.__doc__ = dedent(fun.__doc__) + docstring
        return fun

    return _epilog


def click_run():
    sys.excepthook = exception_handler
    cli(obj={})


@click.group()
def cli():
    pass


@cli.command()
@click.option("--quiet/--no-quiet", default=False)
@click.option(
    "--varformat",
    default="python-format,python-brace-format",
    help=(
        "Comma-separated list of variable formats. " "See Available Variable Formats."
    ),
)
@click.option(
    "--rules",
    default="",
    help=(
        "Comma-separated list of lint rules to use. "
        "Defaults to all rules. See Available Lint Rules."
    ),
)
@click.option(
    "--excluderules",
    default="",
    help=(
        "Comma-separated list of lint rules to exclude. "
        "Defaults to no rules excluded. See Available Lint Rules."
    ),
)
@click.option("--reporter", default="", help="Reporter to use for output.")
@click.option("--errorsonly/--no-errorsonly", default=False, help="Only print errors.")
@click.argument("path", nargs=-1)
@click.pass_context
@epilog(
    format_formats() + "\n" + format_lint_rules() + "\n" + format_lint_template_rules()
)
def lint(ctx, quiet, varformat, rules, excluderules, reporter, errorsonly, path):
    """
    Lints .po/.pot files for issues

    You can ignore rules on a string-by-string basis by adding an
    extracted comment "dennis-ignore: <comma-separated-rules>".  See
    documentation for details.

    """
    if not quiet:
        click.echo(f"dennis version {__version__}")

    # Make sure requested rules are valid
    all_rules = get_linter_rules(with_names=True)
    all_rules.update(get_template_linter_rules(with_names=True))
    rules = [rule.strip() for rule in rules.split(",") if rule.strip()]
    invalid_rules = [rule for rule in rules if rule not in all_rules]
    if invalid_rules:
        raise click.UsageError("invalid rules: %s." % ", ".join(invalid_rules))

    if not rules:
        rules = get_linter_rules()
        rules.update(get_template_linter_rules())
        rules = rules.keys()

    if excluderules:
        excludes = [rule.strip() for rule in excluderules.split(",") if rule.strip()]
        invalid_rules = [rule for rule in excludes if rule not in all_rules]
        if invalid_rules:
            raise click.UsageError(
                "invalid exclude rules: %s." % ", ".join(invalid_rules)
            )

        # Remove excluded rules
        rules = [rule for rule in rules if rule not in excludes]

    # Build linters and lint
    linter = Linter(varformat.split(","), rules)
    templatelinter = TemplateLinter(varformat.split(","), rules)

    po_files = []
    for item in path:
        if os.path.isdir(item):
            for root, dirs, files in os.walk(item):
                po_files.extend(
                    [
                        os.path.join(root, fn)
                        for fn in files
                        if fn.endswith((".po", ".pot"))
                    ]
                )
        else:
            po_files.append(item)

    po_files = [os.path.abspath(fn) for fn in po_files if fn.endswith((".po", ".pot"))]

    if not po_files:
        raise click.UsageError("nothing to work on. Use --help for help.")

    files_to_errors = {}
    total_error_count = 0
    total_warning_count = 0
    total_files_with_errors = 0

    for fn in po_files:
        formatted_fn = click.format_filename(fn)
        try:
            if not os.path.exists(fn):
                raise click.UsageError(f'File "{formatted_fn}" does not exist.')

            if fn.endswith(".po"):
                results = linter.verify_file(fn)
            else:
                results = templatelinter.verify_file(fn)
        except IOError as ioe:
            # This is not a valid .po file. So mark it as an error.
            err(f">>> Problem opening file: {formatted_fn}")
            err(repr(ioe))
            click.echo("")

            # FIXME - should we track this separately as an invalid
            # file?
            files_to_errors[fn] = (1, 0)
            total_error_count += 1
            continue

        if errorsonly:
            # Go through and nix all the non-error LintMessages
            results = [res for res in results if res.kind == "err"]

        # We don't want to print output for files that are fine, so we
        # update the bookkeeping and move on.
        if not results:
            files_to_errors[fn] = (0, 0)
            continue

        if not quiet and not reporter:
            click.secho(f">>> Working on: {formatted_fn}", fg="green", bold=True)

        error_results = [res for res in results if res.kind == "err"]
        warning_results = [res for res in results if res.kind == "warn"]

        error_count = len(error_results)
        total_error_count += error_count

        warning_count = len(warning_results)
        total_warning_count += warning_count

        if not quiet:
            for msg in error_results:
                if reporter == "line":
                    click.echo(f"{fn}: {msg.poentry.linenum}: 0: {msg.code}: {msg.msg}")
                else:
                    err(f"{msg.code}: {msg.msg}")
                    click.echo(withlines(msg.poentry.linenum, msg.poentry.original))
                    click.echo("")

        if not quiet and not errorsonly:
            for msg in warning_results:
                if reporter == "line":
                    click.echo(f"{fn}: {msg.poentry.linenum}: 0: {msg.code}: {msg.msg}")
                else:
                    click.secho(f"{msg.code}: {msg.msg}", fg="yellow", bold=True)
                    click.echo(withlines(msg.poentry.linenum, msg.poentry.original))
                    click.echo("")

        files_to_errors[fn] = (error_count, warning_count)

        if error_count > 0:
            total_files_with_errors += 1

        if not quiet and reporter != "line":
            click.echo("Totals")
            if not errorsonly:
                click.echo(f"  Warnings: {warning_count:5}")
            click.echo(f"  Errors:   {error_count:5}")
            click.echo("")

    if len(po_files) > 1 and not quiet and reporter != "line":
        click.echo("Final totals")
        click.echo(f"  Number of files examined:          {len(po_files):5}")
        click.echo(f"  Total number of files with errors: {total_files_with_errors:5}")
        if not errorsonly:
            click.echo(f"  Total number of warnings:          {total_warning_count:5}")
        click.echo(f"  Total number of errors:            {total_error_count:5}")
        click.echo("")

        file_counts = [
            (counts[0], counts[1], fn.split(os.sep)[-3], fn.split(os.sep)[-1])
            for (fn, counts) in files_to_errors.items()
        ]

        file_counts = list(reversed(sorted(file_counts)))
        printed_header = False
        for error_count, warning_count, locale, fn in file_counts:
            if not error_count and not warning_count:
                continue

            if not printed_header:
                if errorsonly:
                    click.echo("Errors  Filename")
                else:
                    click.echo("Warnings  Errors  Filename")
                printed_header = True

            if errorsonly:
                click.echo(f" {error_count:5}  {locale} ({fn})")
            else:
                click.echo(f"   {warning_count:5}   {error_count:5}  {locale} ({fn})")

    # Return 0 if everything was fine or 1 if there were errors.
    ctx.exit(code=1 if total_error_count else 0)


@cli.command()
@click.option(
    "--showuntranslated", is_flag=True, default=False, help="Show untranslated strings"
)
@click.option("--showfuzzy", is_flag=True, default=False, help="Show fuzzy strings")
@click.argument("path", nargs=-1, type=click.Path(exists=True))
@click.pass_context
def status(ctx, showuntranslated, showfuzzy, path):
    """Show status of a .po file."""
    click.echo(f"dennis version {__version__}")

    po_files = []
    for item in path:
        if os.path.isdir(item):
            for root, dirs, files in os.walk(item):
                po_files.extend(
                    [os.path.join(root, fn) for fn in files if fn.endswith(".po")]
                )
        else:
            po_files.append(item)

    po_files = [os.path.abspath(fn) for fn in po_files if fn.endswith(".po")]

    for fn in po_files:
        formatted_fn = click.format_filename(fn)
        try:
            if not os.path.exists(fn):
                raise IOError(f'File "{formatted_fn}" does not exist.')

            pofile = parse_pofile(fn)
        except IOError as ioe:
            err(f">>> Problem opening file: {formatted_fn}")
            err(repr(ioe))
            continue

        click.echo("")
        click.secho(f">>> Working on: {formatted_fn}", fg="green", bold=True)

        click.echo("Metadata:")
        for key in (
            "Language",
            "Report-Msgid-Bugs-To",
            "PO-Revision-Date",
            "Last-Translator",
            "Language-Team",
            "Plural-Forms",
        ):
            if key in pofile.metadata and pofile.metadata[key]:
                click.echo(f"  {key}: {pofile.metadata[key]}")
        click.echo("")

        if showuntranslated:
            click.echo("Untranslated strings:")
            click.echo("")
            for poentry in pofile.untranslated_entries():
                click.echo(withlines(poentry.linenum, poentry.original))
                click.echo("")

        if showfuzzy:
            click.echo("Fuzzy strings:")
            click.echo("")
            for poentry in pofile.fuzzy_entries():
                click.echo(withlines(poentry.linenum, poentry.original))
                click.echo("")

        total_words = 0
        translated_words = 0
        untranslated_words = 0
        for entry in pofile.translated_entries() + pofile.untranslated_entries():
            if not entry.translated():
                untranslated_words += len(entry.msgid.split())
            else:
                translated_words += len(entry.msgid.split())
            total_words += len(entry.msgid.split())

        total_strings = len(pofile)
        translated_total = len(pofile.translated_entries())
        fuzzy_total = len(pofile.fuzzy_entries())
        untranslated_total = len(pofile.untranslated_entries())

        click.echo("Statistics:")
        click.echo(f"  Total strings:             {total_strings}")
        click.echo(f"    Translated:              {translated_total}")
        click.echo(f"    Untranslated:            {untranslated_total}")
        click.echo(f"    Fuzzy:                   {fuzzy_total}")
        click.echo(f"  Total translateable words: {total_words}")
        click.echo(f"    Translated:              {translated_words}")
        click.echo(f"    Untranslated:            {untranslated_words}")
        if untranslated_words == 0:
            click.echo("  Percentage:                100% COMPLETE!")
        else:
            click.echo(f"  Percentage:                {pofile.percent_translated()}%")

    ctx.exit(0)


@cli.command()
@click.option(
    "--varformat",
    default="python-format,python-brace-format",
    help=("Comma-separated list of variable types. " "See Available Variable Formats."),
)
@click.option(
    "--pipeline",
    "-p",
    default="pirate",
    help=("Comma-separated translate pipeline. See Available " "Pipeline Parts."),
)
@click.option(
    "--strings",
    "-s",
    default=False,
    is_flag=True,
    help="Command line args are strings to be translated",
)
@click.argument("path", nargs=-1)
@click.pass_context
@epilog(format_formats() + "\n" + format_pipeline_parts())
def translate(ctx, varformat, pipeline, strings, path):
    """
    Translate a single string or .po file of strings.

    If you want to pull the string from stdin, use "-".

    Note: Translating files is done in-place replacing the original
    file.

    """
    if not (path and path[0] == "-"):
        # We don't want to print this if they're piping to stdin
        click.echo(f"dennis version {__version__}")

    if not path:
        raise click.UsageError("nothing to work on. Use --help for help.")

    try:
        translator = Translator(varformat.split(","), pipeline.split(","))
    except InvalidPipeline as ipe:
        raise click.UsageError(ipe.args[0])

    if strings:
        # Args are strings to be translated
        for arg in path:
            data = translator.translate_string(arg)
            click.echo(data)

    elif path[0] == "-":
        # Read everything from stdin, then translate it
        data = click.get_binary_stream("stdin").read()
        data = translator.translate_string(data)
        click.echo(data)

    else:
        # Check all the paths first
        for arg in path:
            if not os.path.exists(arg):
                raise click.UsageError(
                    "File {fn} does not exist.".format(fn=click.format_filename(arg))
                )

        for arg in path:
            click.echo(translator.translate_file(arg))

    ctx.exit(0)


def exception_handler(exc_type, exc_value, exc_tb):
    click.echo("Oh no! Dennis has thrown an error while trying to do stuff.")
    click.echo("Please write up a bug report with the specifics so that ")
    click.echo("we can fix it.")
    click.echo("")
    click.echo("https://github.com/mozilla/dennis/issues")
    click.echo("")
    click.echo("Here is some information you can copy and paste into the ")
    click.echo("bug report:")
    click.echo("")
    click.echo("---")
    click.echo(f"Dennis: {__version__}")
    click.echo(f"Python: {sys.version}")
    click.echo(f"Command line: {sys.argv}")
    click.echo("".join(traceback.format_exception(exc_type, exc_value, exc_tb)))
    click.echo("---")



import re

import click


class Format:
    """Variable format base class"""

    name = ""
    desc = ""
    regexp = ""

    identifier = None

    @classmethod
    def extract_variable_name(cls, text):
        raise NotImplementedError


# https://www.gnu.org/software/gettext/manual/html_node/python_002dformat.html#python_002dformat
# https://docs.python.org/2/library/string.html#format-string-syntax
class PythonBraceFormat(Format):
    name = "python-brace-format"
    desc = 'Python brace format (e.g. "{0}", "{foo}")'

    regexp = (
        # {}, {0}, {foo}, {foo:bar}, {foo:bar baz}
        r"(?:\{[^\}]*?\})"
    )

    identifier = re.compile(r"\{([^!:\}]*)")

    @classmethod
    def extract_variable_name(cls, text):
        identifier = cls.identifier.match(text)
        if identifier:
            return identifier.group(1)


# https://www.gnu.org/software/gettext/manual/html_node/python_002dformat.html#python_002dformat
# https://docs.python.org/2/library/stdtypes.html#string-formatting-operations
class PythonFormat(Format):
    name = "python-format"
    desc = 'Python percent format (e.g. "%s", "%(foo)s")'

    regexp = (
        # %s and %(foo)s
        # Note: This doesn't support %E or %F because of problems
        # with false positives and urlencoding. Theoretically those
        # aren't getting used in gettext contexts anyhow.
        r"(?:%(?:[(]\S+?[)])?[#0+-]?[\.\d\*]*[hlL]?[diouxefGgcrs])"
    )

    identifier = re.compile(
        r"%" r"(?:" + r"\((\S+?)\)" + r")?" r"[#0+-]?[\.\d\*]*[hlL]?[diouxefGgcrs]"
    )

    @classmethod
    def extract_variable_name(cls, text):
        identifier = cls.identifier.match(text)
        if identifier:
            return identifier.group(1) or ""


def get_available_formats():
    return {
        thing.name: thing
        for name, thing in globals().items()
        if (name.endswith("Format") and issubclass(thing, Format) and thing.name)
    }


class UnknownFormat(Exception):
    pass


class VariableTokenizer:
    def __init__(self, formats=None):
        """
        :arg formats: List of variable formats

            If None, creates a VariableTokenizer that tokenizes on all
            formats of variables. Otherwise just recognizes the listed
            formats.

        """
        all_formats = get_available_formats()

        if formats is None:
            formats = all_formats.keys()

        formats = [fmt for fmt in formats if fmt]

        # If they don't want variable tokenizing at all
        if not formats:
            self.formats = []
            self.vars_re = None

        else:
            # Convert names to classes
            self.formats = []

            for fmt in formats:
                try:
                    self.formats.append(all_formats[fmt])
                except KeyError:
                    raise UnknownFormat("{} is not a known variable format".format(fmt))

            # Generate variable regexp
            self.vars_re = re.compile(
                r"(" + "|".join([vt.regexp for vt in self.formats]) + r")"
            )

    def contains(self, fmt):
        """Does this tokenizer contain specified variable format?"""
        return fmt in [tok.name for tok in self.formats]

    def tokenize(self, text):
        """Breaks s into strings and Python variables

        This preserves whitespace.

        :arg text: the string to tokenize

        :returns: list of tokens---every even one is a Python variable

        """
        if not self.vars_re:
            return [text]
        return [token for token in self.vars_re.split(text) if token]

    def extract_tokens(self, text, unique=True):
        """Returns the set of variable in the text"""
        if not self.vars_re:
            return set()

        try:
            tokens = self.vars_re.findall(text)
            if unique:
                tokens = set(tokens)
            return tokens
        except TypeError:
            click.echo("TYPEERROR: {}".format(repr(text)))

    def is_token(self, text):
        """Is this text a variable?"""
        if not self.vars_re:
            return False
        return self.vars_re.match(text) is not None

    def extract_variable_name(self, text):
        for fmt in self.formats:
            if re.compile(fmt.regexp).match(text):
                return fmt.extract_variable_name(text)


def all_subclasses(cls):
    subc = cls.__subclasses__()
    for d in list(subc):
        subc.extend(all_subclasses(d))
    return subc


# Matches:
# "dennis-ignore: *" to skip all the rules
# "dennis-ignore: E201,..." to ignore specific rules
DENNIS_NOTE_RE = re.compile(r"dennis-ignore:\s+(\*|[EW0-9,]+)")


def parse_dennis_note(text):
    """Parses a dennis note and returns list of rules to skip"""
    if not text:
        return []

    match = DENNIS_NOTE_RE.search(text)
    if not match:
        return []

    match = match.group(1).strip()
    if match == "*":
        return "*"

    return [item for item in match.split(",") if item]


def parse_pofile(fn_or_string):
    """Parses a po file and attaches original poentry blocks

    When polib parses a pofile, it captures the line number of the
    start of the block, but doesn't capture the original string for
    the block. When you call str() on the poentry, it "reassembles"
    the block with textwrapped lines, so it returns something
    substantially different than the original block. This is
    problematic if we want to print out the block with the line
    numbers--one for each line.

    So this wrapper captures the line numbers and original text for
    each block and attaches that to the parsed poentries in an
    attribute named "original" thus allowing us to print the original
    text with line numbers.

    """
    from polib import _is_file, detect_encoding, io, pofile

    # This parses the pofile
    parsed_pofile = pofile(fn_or_string)

    # Now we need to build a linenumber -> block hash so that we can
    # accurately print out what was in the pofile because polib will
    # reassembled what it parsed, but it's not the same.
    if _is_file(fn_or_string):
        enc = detect_encoding(fn_or_string, "pofile")
        fp = io.open(fn_or_string, "rt", encoding=enc)
    else:
        fp = fn_or_string.splitlines(True)

    fp = list(fp)
    entries = list(parsed_pofile)
    for i, poentry in enumerate(entries):
        # Grab the lines that make up the poentry.
        # Note: linenum is 1-based, so we convert it to 0-based.
        try:
            lines = fp[poentry.linenum - 1 : entries[i + 1].linenum - 1]
        except IndexError:
            lines = fp[poentry.linenum - 1 :]

        # Nix blank lines at the end.
        while lines and not lines[-1].strip():
            lines.pop()

        # Join them and voila!
        poentry.original = "".join(lines)

    return parsed_pofile


def withlines(linenum, poentry_text):
    """Returns text with line numbers"""
    start = linenum
    new_text = []

    lines_with_nums = zip(range(start, start + 100), poentry_text.splitlines())

    for line_no, line in lines_with_nums:
        new_text.append(str(line_no) + ":" + str(line))

    return "\n".join(new_text)



from dennis.tools import (
    VariableTokenizer,
    all_subclasses,
    parse_dennis_note,
    parse_pofile,
)
from dennis.linter import LintedEntry, LintMessage

WARNING = "warn"
ERROR = "err"


class TemplateLintRule:
    num = ""
    name = ""
    desc = ""

    def lint(self, vartok, linted_entry):
        """Takes a linted entry and adds errors and warnings

        :arg vartok: the variable tokenizer to use for tokenizing
            on variable tokens
        :arg linted_entry: the LintedEntry to work on

        """
        raise NotImplementedError


class HardToReadNamesTLR(TemplateLintRule):
    num = "W500"
    name = "hardtoread"
    desc = "Looks for vars that are hard to read like o, O, 0, l, 1"

    hard_to_read = ("o", "O", "0", "l", "1")

    def lint(self, vartok, linted_entry):
        msgs = []

        idstr = linted_entry.str

        for s in idstr.msgid_strings:
            if not s:
                continue

            msgid_tokens = vartok.extract_tokens(s)
            msgid_tokens = [
                vartok.extract_variable_name(token) for token in msgid_tokens
            ]

            for token in msgid_tokens:
                if token in self.hard_to_read:
                    msgs.append(
                        LintMessage(
                            WARNING,
                            linted_entry.poentry.linenum,
                            0,
                            self.num,
                            'hard to read variable name "{}"'.format(token),
                            linted_entry.poentry,
                        )
                    )
        return msgs


class OneCharNamesTLR(TemplateLintRule):
    num = "W501"
    name = "onechar"
    desc = "Looks for one character variable names"

    def lint(self, vartok, linted_entry):
        msgs = []
        idstr = linted_entry.str

        for s in idstr.msgid_strings:
            if not s:
                continue

            msgid_tokens = vartok.extract_tokens(s)
            msgid_tokens = [
                vartok.extract_variable_name(token) for token in msgid_tokens
            ]

            for token in msgid_tokens:
                if len(token) == 1 and token.isalpha():
                    msgs.append(
                        LintMessage(
                            WARNING,
                            linted_entry.poentry.linenum,
                            0,
                            self.num,
                            'one character variable name "{}"'.format(token),
                            linted_entry.poentry,
                        )
                    )
        return msgs


class MultipleUnnamedVarsTLR(TemplateLintRule):
    num = "W502"
    name = "unnamed"
    desc = "Looks for multiple unnamed variables"

    def lint(self, vartok, linted_entry):
        msgs = []
        idstr = linted_entry.str

        for s in idstr.msgid_strings:
            if not s:
                continue

            msgid_tokens = vartok.extract_tokens(s, unique=False)
            msgid_tokens = [
                vartok.extract_variable_name(token) for token in msgid_tokens
            ]

            if msgid_tokens.count("") > 1:
                msgs.append(
                    LintMessage(
                        WARNING,
                        linted_entry.poentry.linenum,
                        0,
                        self.num,
                        "multiple variables with no name.",
                        linted_entry.poentry,
                    )
                )
        return msgs


def get_lint_rules(with_names=False):
    lint_rules = {}

    for cls in all_subclasses(TemplateLintRule):
        if cls.num:
            lint_rules[cls.num] = cls
            if with_names and cls.name:
                lint_rules[cls.name] = cls

    return lint_rules


def convert_rules(rules_spec):
    lint_rules = get_lint_rules(with_names=True)
    rules = [lint_rules[rule]() for rule in rules_spec if rule in lint_rules]
    return rules


class TemplateLinter:
    def __init__(self, vars_, rules_spec):
        self.vartok = VariableTokenizer(vars_)
        self.rules_spec = rules_spec
        self.rules = convert_rules(self.rules_spec)

    def lint_poentry(self, poentry):
        linted_entry = LintedEntry(poentry)

        skip = parse_dennis_note(poentry.comment)

        msgs = []

        # Check the comment to see if what we should ignore.
        for lint_rule in self.rules:
            if skip == "*" or lint_rule.num in skip:
                continue

            msgs.extend(lint_rule.lint(self.vartok, linted_entry))

        return msgs

    def verify_file(self, filename_or_string):
        """Verifies strings in file.

        :arg filename_or_string: filename to verify or the contents of
            a pofile as a string

        :returns: list of LintMessage objects

        :raises IOError: if the file is not a valid .po file or
            doesn't exist
        """
        po = parse_pofile(filename_or_string)
        msgs = []
        for entry in po:
            msgs.extend(self.lint_poentry(entry))
        return msgs



import polib

from dennis.linter import (
    BadFormatLintRule,
    BlankLintRule,
    MalformedNoTypeLintRule,
    MalformedMissingRightBraceLintRule,
    MalformedMissingLeftBraceLintRule,
    MismatchedHTMLLintRule,
    MissingVarsLintRule,
    InvalidVarsLintRule,
    UnchangedLintRule,
    LintedEntry,
    Linter,
)
from dennis.templatelinter import (
    HardToReadNamesTLR,
    OneCharNamesTLR,
    MultipleUnnamedVarsTLR,
)
from dennis.tools import VariableTokenizer
from tests import build_po_string


class LinterTest:
    def test_linter(self):
        pofile = build_po_string("#: foo/foo.py:5\n" 'msgid "Foo"\n' 'msgstr "Oof"\n')

        linter = Linter(["python-format", "python-brace-format"], ["E201", "W202"])
        msgs = linter.verify_file(pofile)

        # No warnings or errors, so there are no messages.
        assert len(msgs) == 0

    def test_linter_fuzzy_strings(self):
        pofile = build_po_string(
            "#, fuzzy\n"
            '#~ msgid "Most Recent Message"\n'
            '#~ msgid_plural "Last %(count)s Messages"\n'
            '#~ msgstr[0] "Les {count} derniers messages"\n'
            '#~ msgstr[1] "Les {count} derniers messages"\n'
        )

        linter = Linter(["python-format", "python-brace-format"], ["E201", "W202"])
        msgs = linter.verify_file(pofile)

        # There were no non-fuzzy strings, so nothing to lint.
        assert len(msgs) == 0

    def test_linter_untranslated_strings(self):
        pofile = build_po_string(
            "#, fuzzy\n"
            '#~ msgid "Most Recent Message"\n'
            '#~ msgid_plural "Last %(count)s Messages"\n'
            '#~ msgstr[0] ""\n'
            '#~ msgstr[1] ""\n'
        )

        linter = Linter(["python-format", "python-brace-format"], ["E201", "W202"])
        msgs = linter.verify_file(pofile)

        # There were no translated strings, so nothing to lint.
        assert len(msgs) == 0


def build_linted_entry(po_data):
    po = polib.pofile(build_po_string(po_data))
    poentry = list(po)[0]
    return LintedEntry(poentry)


class LintRuleTestCase:
    vartok = VariableTokenizer(["python-format", "python-brace-format"])


class TestBadFormatLintRule(LintRuleTestCase):
    lintrule = BadFormatLintRule()

    def test_fine(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo"\n' 'msgstr "FOO"\n'
        )
        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert msgs == []

        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo %s"\n' 'msgstr "FOO %s"\n'
        )
        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert msgs == []

    def test_bad_format_character(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "%s foo"\n' 'msgstr "%a FOO"\n'
        )
        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "err"
        assert msgs[0].code == "E104"
        assert msgs[0].msg == "bad format character: %a"

    def test_last_character(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "foo %s"\n' 'msgstr "FOO %"\n'
        )
        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "err"
        assert msgs[0].code == "E104"
        assert msgs[0].msg == "bad format character: %"

    def test_handle_double_percent(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "%% foo"\n' 'msgstr "%% FOO"\n'
        )
        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert msgs == []

    def test_not_in_msgid(self):
        # If there are no vars in the msgid, then we don't want to throw errors about stuff we see
        # in the translated string because it's likely it's not being interpolated by %.
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "foo 50% omg"\n' 'msgstr "FOO 50% OMG"\n'
        )
        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert msgs == []

    def test_ignore_non_formatting_tokens(self):
        # If the formatting token is actually part of another formatting token, then we want to
        # ignore it.
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n"
            'msgid "foo {startdate:%Y-%m-%d %H:%M} bar"\n'
            'msgstr "FOO {startdate:%Y-%m-%d %H:%M} BAR"\n'
        )
        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert msgs == []

    def test_varformat_empty(self):
        vartok = VariableTokenizer([])
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "%s foo"\n' 'msgstr "%a FOO"\n'
        )
        msgs = self.lintrule.lint(vartok, linted_entry)
        assert msgs == []


class TestMalformedNoTypeLintRule(LintRuleTestCase):
    lintrule = MalformedNoTypeLintRule()

    def test_fine(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo"\n' 'msgstr "Oof"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 0

        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo: {foo}"\n' 'msgstr "Oof: {foo}"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert msgs == []

    def test_python_var_with_space(self):
        linted_entry = build_linted_entry(
            "#: kitsune/questions/templates/questions/answers.html:56\n"
            'msgid "%(count)s view"\n'
            'msgid_plural "%(count)s views"\n'
            'msgstr[0] "%(count) zoo"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "err"
        assert msgs[0].code == "E101"
        assert msgs[0].msg == "type missing: %(count)"

    def test_python_var_end_of_line(self):
        linted_entry = build_linted_entry(
            "#: kitsune/questions/templates/questions/answers.html:56\n"
            'msgid "%(count)s"\n'
            'msgid_plural "%(count)s"\n'
            'msgstr[0] "%(count)"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "err"
        assert msgs[0].code == "E101"
        assert msgs[0].msg == "type missing: %(count)"

    def test_python_var_punctuation(self):
        linted_entry = build_linted_entry(
            "#: kitsune/questions/templates/questions/answers.html:56\n"
            'msgid "%(count)s"\n'
            'msgstr "%(count)!"\n'
        )
        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "err"
        assert msgs[0].code == "E101"
        assert msgs[0].msg == "type missing: %(count)"

    def test_python_var_not_malformed(self):
        """This used to be a false positive"""
        linted_entry = build_linted_entry(
            "#: kitsune/questions/templates/questions/answers.html:56\n"
            'msgid "%(stars)s by %(user)s on %(date)s (%(locale)s)"\n'
            'msgstr "%(stars)s de %(user)s el %(date)s (%(locale)s)"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert msgs == []

    def test_varformat_empty(self):
        vartok = VariableTokenizer([])

        linted_entry = build_linted_entry(
            "#: kitsune/questions/templates/questions/answers.html:56\n"
            'msgid "%(count)s view"\n'
            'msgid_plural "%(count)s views"\n'
            'msgstr[0] "%(count) zoo"\n'
        )

        msgs = self.lintrule.lint(vartok, linted_entry)
        assert msgs == []


class TestMalformedMissingRightBraceLintRule(LintRuleTestCase):
    lintrule = MalformedMissingRightBraceLintRule()

    def test_python_var_missing_right_curly_brace(self):
        linted_entry = build_linted_entry(
            "#: kitsune/questions/templates/questions/answers.html:56\n"
            'msgid "{foo} bar is the best thing ever"\n'
            'msgstr "{foo) bar is the best thing ever"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "err"
        assert msgs[0].code == "E102"
        assert (
            msgs[0].msg == "missing right curly-brace: {foo) bar is the best thing ever"
        )

        linted_entry = build_linted_entry(
            "#: kitsune/questions/templates/questions/answers.html:56\n"
            'msgid "{foo}"\n'
            'msgstr "{foo"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "err"
        assert msgs[0].code == "E102"
        assert msgs[0].msg == "missing right curly-brace: {foo"

    def test_python_var_missing_right_curly_brace_two_vars(self):
        # Test right-most one
        linted_entry = build_linted_entry(
            'msgid "Value for key \\"{0}\\" exceeds the length of {1}"\n'
            'msgstr "Valor para la clave \\"{0}\\" excede el tamano de {1]"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "err"
        assert msgs[0].code == "E102"
        assert msgs[0].msg == "missing right curly-brace: {1]"

        # Test left-most one
        linted_entry = build_linted_entry(
            'msgid "Value for key \\"{0}\\" exceeds the length of {1}"\n'
            'msgstr "Valor para la clave \\"{0]\\" excede el tamano de {1}"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "err"
        assert msgs[0].code == "E102"
        assert msgs[0].msg == 'missing right curly-brace: {0]" excede el tamano de {'

        linted_entry = build_linted_entry(
            "#: kitsune/questions/templates/questions/question_details.html:14\n"
            'msgid "{q} | {product} Support Forum"\n'
            'msgstr "{q} | {product}} foo bar"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "err"
        assert msgs[0].code == "E102"
        assert msgs[0].msg == "missing right curly-brace: {product}} foo bar"

    def test_varformat_empty(self):
        vartok = VariableTokenizer([])

        linted_entry = build_linted_entry(
            'msgid "Value for key \\"{0}\\" exceeds the length of {1}"\n'
            'msgstr "Valor para la clave \\"{0}\\" excede el tamano de {1]"\n'
        )

        msgs = self.lintrule.lint(vartok, linted_entry)
        assert msgs == []


class TestMalformedMissingLeftBraceLintRuleTest(LintRuleTestCase):
    lintrule = MalformedMissingLeftBraceLintRule()

    def test_python_var_missing_left_curly_brace(self):
        linted_entry = build_linted_entry(
            "#: kitsune/questions/templates/questions/answers.html:56\n"
            'msgid "{product} Support Forum"\n'
            'msgstr "product}-Hilfeforum"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "err"
        assert msgs[0].code == "E103"
        assert msgs[0].msg == "missing left curly-brace: product}"

        linted_entry = build_linted_entry(
            "#: kitsune/questions/templates/questions/answers.html:56\n"
            'msgid "{q} | {product} Support Forum"\n'
            'msgstr "{q} | product}-Hilfeforum"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "err"
        assert msgs[0].code == "E103"
        assert msgs[0].msg == "missing left curly-brace: } | product}"

    def test_double_braces(self):
        linted_entry = build_linted_entry(
            'msgid "This is {{literal}} brace, {0}, and {{another}}."\n'
            'msgstr "This is {{literal}} brace, {0}, and {{another}}."\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 0

    def test_varformat_empty(self):
        vartok = VariableTokenizer([])

        linted_entry = build_linted_entry(
            "#: kitsune/questions/templates/questions/question_details.html:14\n"
            'msgid "{q} | {product} Support Forum"\n'
            'msgstr "{q} | {product}} foo bar"\n'
        )

        msgs = self.lintrule.lint(vartok, linted_entry)
        assert msgs == []


class TestMissingVarsLintRule(LintRuleTestCase):
    lintrule = MissingVarsLintRule()

    def test_fine(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo"\n' 'msgstr "Oof"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 0

        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo: {foo}"\n' 'msgstr "Oof: {foo}"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 0

    def test_missing(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo: {foo}"\n' 'msgstr "Oof"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "warn"
        assert msgs[0].code == "W202"
        assert msgs[0].msg == "missing variables: {foo}"

        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n"
            'msgid "Foo: {foo} {bar} {baz}"\n'
            'msgstr "Oof: {foo}"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "warn"
        assert msgs[0].code == "W202"
        assert msgs[0].msg == "missing variables: {bar}, {baz}"

    def test_complex(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n"
            'msgid "Foo: {bar} {baz}"\n'
            'msgstr "Oof: {foo} {bar}"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "warn"
        assert msgs[0].code == "W202"
        assert msgs[0].msg == "missing variables: {baz}"

    def test_plurals(self):
        # It's possible for the msgid to have no variables in it and
        # the msgid_plural to have variables in it. msgstr[n] strings
        # should be compared against the set of variables in msgid and
        # msgid_plural. This tests the common case.
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n"
            'msgid "1 reply"\n'
            'msgid_plural "{n} replies"\n'
            'msgstr[0] "{n} mooo"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 0

    def test_plurals_not_missing(self):
        # If the msgstr doesn't have variables that are in msgid or
        # msgid_plural, that's ok since we don't know which plurality
        # it is.
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n"
            'msgid "1 post"\n'
            'msgid_plural "{0} posts"\n'
            'msgstr[0] "1 moo"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 0

    def test_double_percent(self):
        # Double-percent shouldn't be picked up as a variable.
        # Issue #28.
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n"
            'msgid "50% of the traffic"\n'
            'msgstr "more than 50%% of the traffic"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 0

    def test_urlencoded_urls(self):
        # urlencoding uses % and that shouldn't get picked up
        # as variables.
        # Issue #27.
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n"
            'msgid "OMG! Best url is http://example.com/foo"\n'
            'msgstr "http://example.com/foo%20%E5%B4%A9%E6%BA%83 is best"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 0

    def test_python_format_are_errors_unnamed(self):
        # python-format unnamed variables must be in the msgstr so they're errors.
        # Issue #57
        linted_entry = build_linted_entry(
            "#: kitsune/kbforums/feeds.py:23\n"
            'msgid "Recently updated threads about %s"\n'
            'msgstr "RECENTLY UPDATED THREADS"\n'
        )
        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "err"
        assert msgs[0].code == "E202"
        assert msgs[0].msg == "missing variables: %s"

    def test_varformat_empty(self):
        vartok = VariableTokenizer([])

        linted_entry = build_linted_entry(
            "#: kitsune/kbforums/feeds.py:23\n"
            'msgid "Recently updated threads about %s"\n'
            'msgstr "RECENTLY UPDATED THREADS"\n'
        )
        msgs = self.lintrule.lint(vartok, linted_entry)
        assert msgs == []


class TestInvalidVarsLintRule(LintRuleTestCase):
    lintrule = InvalidVarsLintRule()

    def test_fine(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo"\n' 'msgstr "Oof"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert msgs == []

        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo: {foo}"\n' 'msgstr "Oof: {foo}"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert msgs == []

    def test_invalid(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo {bar}"\n' 'msgstr "Oof: {foo}"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "err"
        assert msgs[0].code == "E201"
        assert msgs[0].msg == "invalid variables: {foo}"

        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n"
            'msgid "Foo: {foo}"\n'
            'msgstr "Oof: {foo} {bar} {baz}"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "err"
        assert msgs[0].code == "E201"
        assert msgs[0].msg == "invalid variables: {bar}, {baz}"

    def test_complex(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n"
            'msgid "Foo: {bar} {baz}"\n'
            'msgstr "Oof: {foo} {bar}"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "err"
        assert msgs[0].code == "E201"
        assert msgs[0].msg == "invalid variables: {foo}"

    def test_plurals(self):
        # It's possible for the msgid to have no variables in it and
        # the msgid_plural to have variables in it. msgstr[n] strings
        # should be compared against the set of variables in msgid and
        # msgid_plural. This tests the common case.
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n"
            'msgid "1 reply"\n'
            'msgid_plural "{n} replies"\n'
            'msgstr[0] "{n} mooo"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert msgs == []

    def test_double_percent(self):
        # Double-percent shouldn't be picked up as a variable.
        # Issue #28.
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n"
            'msgid "50% of the traffic"\n'
            'msgstr "more than 50%% of the traffic"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert msgs == []

    def test_urlencoded_urls(self):
        # urlencoding uses % and that shouldn't get picked up
        # as variables.
        # Issue #27.
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n"
            'msgid "OMG! Best url is http://example.com/foo"\n'
            'msgstr "http://example.com/foo%20%E5%B4%A9%E6%BA%83 is best"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert msgs == []

    def test_msgid_no_vars(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n"
            'msgid "http://en.wikipedia.org/wiki/Canvas_element"\n'
            'msgstr "http://it.wikipedia.org/wiki/Canvas_%28elemento_HTML%29"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert msgs == []

    def test_varformat_empty(self):
        vartok = VariableTokenizer([])
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo {bar}"\n' 'msgstr "Oof: {foo}"\n'
        )

        msgs = self.lintrule.lint(vartok, linted_entry)
        assert msgs == []


class TestBlankLintRule(LintRuleTestCase):
    lintrule = BlankLintRule()

    def test_fine(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo"\n' 'msgstr ""\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 0

    def test_whitespace(self):
        testdata = [" ", "  ", "\t"]
        for data in testdata:
            linted_entry = build_linted_entry(
                "#: foo/foo.py:5\n" 'msgid "Foo"\n' 'msgstr "%s"\n' % data
            )

            msgs = self.lintrule.lint(self.vartok, linted_entry)
            assert len(msgs) == 1
            assert msgs[0].kind == "warn"
            assert msgs[0].code == "W301"
            assert msgs[0].msg == "translated string is solely whitespace"


class TestUnchangedLintRule(LintRuleTestCase):
    lintrule = UnchangedLintRule()

    def test_unchanged(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo"\n' 'msgstr "Foo"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "warn"
        assert msgs[0].code == "W302"
        assert msgs[0].msg == "translated string is same as source string"


class TestMismatchedHTMLLintRule(LintRuleTestCase):
    lintrule = MismatchedHTMLLintRule()

    def test_fine(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "<b>Foo</b>"\n' 'msgstr "<b>ARGH</b>"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 0

    def test_fail(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "<b>Foo</b>"\n' 'msgstr "<em>ARGH</em>"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "warn"
        assert msgs[0].code == "W303"
        assert msgs[0].msg == 'different html: "</b>" vs. "</em>"'

    def test_different_numbers(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "<b>Foo"\n' 'msgstr "<b>ARGH</b>"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "warn"
        assert msgs[0].code == "W303"
        assert msgs[0].msg == 'different html: "<b>" vs. "</b>"'

        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "<b>Foo</b>"\n' 'msgstr "<b>ARGH"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "warn"
        assert msgs[0].code == "W303"
        assert msgs[0].msg == 'different html: "</b>" vs. "<b>"'

    def test_invalid_html(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" + 'msgid "<a>Foo</a>"\n' + 'msgstr "<a>ARGH</\u0430>"\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        assert msgs[0].kind == "warn"
        assert msgs[0].code == "W303"
        # HTMLParser doesn't recognize </\u0430> as a valid HTML tag,
        # so you end up with:
        #
        # [<html </a>>, <html <a>>]
        #
        # vs.
        #
        # [<html <a>>]
        assert msgs[0].msg == 'different html: "</a>" vs. "<a>"'


class TLRTestCase:
    vartok = VariableTokenizer(["python-format", "python-brace-format"])


class TestHardToReadNamesTLR(TLRTestCase):
    lintrule = HardToReadNamesTLR()

    def test_hard_to_read_names(self):
        for c in ("o", "O", "0", "l", "1"):
            linted_entry = build_linted_entry(
                "#: foo/foo.py:5\n" 'msgid "Foo: %(' + c + ')s"\n' 'msgstr ""\n'
            )

            msgs = self.lintrule.lint(self.vartok, linted_entry)
            assert len(msgs) == 1

            linted_entry = build_linted_entry(
                "#: foo/foo.py:5\n" 'msgid "Foo: {' + c + '}"\n' 'msgstr ""\n'
            )

            msgs = self.lintrule.lint(self.vartok, linted_entry)
            assert len(msgs) == 1
        # FIXME: flesh out this test


class TestMultipleUnnamedVarsTLR(TLRTestCase):
    lintrule = MultipleUnnamedVarsTLR()

    def test_multi_vars_no_name(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo: %s %s"\n' 'msgstr ""\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1

        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo: {} {}"\n' 'msgstr ""\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        # FIXME: flesh out this test


class TestOneCharNamesTLR(TLRTestCase):
    lintrule = OneCharNamesTLR()

    def test_one_character_names(self):
        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo: %(c)s"\n' 'msgstr ""\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1

        linted_entry = build_linted_entry(
            "#: foo/foo.py:5\n" 'msgid "Foo: {c}"\n' 'msgstr ""\n'
        )

        msgs = self.lintrule.lint(self.vartok, linted_entry)
        assert len(msgs) == 1
        # FIXME: flesh out this test



from textwrap import dedent

from click.testing import CliRunner
import pytest

from dennis.cmdline import cli
from tests import build_po_string, nix_header


@pytest.fixture
def runner():
    """Creates a CliRunner"""
    return CliRunner()


def build_key_val(text):
    pairs = {}
    for line in text.splitlines():
        if ":" in line:
            key, val = line.split(":", 1)
            pairs[key.strip()] = val.strip()
    return pairs


class TestStatus:
    def test_help(self, runner):
        result = runner.invoke(cli, ("status", "--help"))
        assert result.exit_code == 0
        assert "Usage" in result.output.splitlines()[0]

    def test_status_untranslated(self, runner, tmpdir):
        po_file = build_po_string(
            "#: foo/foo.py:5\n" 'msgid "Foo bar baz"\n' 'msgstr ""\n'
        )
        fn = tmpdir.join("messages.po")
        fn.write(po_file)

        result = runner.invoke(cli, ("status", str(fn)))
        assert result.exit_code == 0

        # FIXME: This would be a lot easier if the status was structured and parseable.
        pairs = build_key_val(result.output)
        assert pairs["Total strings"] == "1"
        assert pairs["Total translateable words"] == "3"
        assert pairs["Percentage"] == "0%"

    def test_status_translated(self, runner, tmpdir):
        po_file = build_po_string(
            "#: foo/foo.py:5\n" 'msgid "Foo bar baz"\n' 'msgstr "Feh"\n'
        )
        fn = tmpdir.join("messages.po")
        fn.write(po_file)

        result = runner.invoke(cli, ("status", str(fn)))
        assert result.exit_code == 0

        # FIXME: This would be a lot easier if the status was structured and parseable.
        pairs = build_key_val(result.output)
        assert pairs["Total strings"] == "1"
        assert pairs["Total translateable words"] == "3"
        assert pairs["Percentage"] == "100% COMPLETE!"

    # FIXME: test --showuntranslated on .po file

    # FIXME: test --showfuzzy on .po file


class TestTranslate:
    def test_help(self, runner):
        result = runner.invoke(cli, ("translate", "--help"))
        assert result.exit_code == 0
        assert "Usage" in result.output.splitlines()[0]

    def test_strings(self, runner):
        result = runner.invoke(cli, ("translate", "-p", "shouty", "-s", "ou812"))
        assert result.exit_code == 0
        assert "OU812" in result.output

    def test_no_paths(self, runner):
        result = runner.invoke(cli, ("translate", "-p", "shouty"))
        assert result.exit_code == 2
        last_line = result.output.splitlines()[-1]
        assert last_line == "Error: nothing to work on. Use --help for help."

    def test_pipe(self, runner):
        result = runner.invoke(cli, ("translate", "-p", "shouty", "-"), input="foo bar")
        assert result.exit_code == 0
        assert result.output == "FOO BAR\n"

    def test_pipeline(self, runner, tmpdir):
        po_file = build_po_string(
            "#: foo/foo.py:5\n" 'msgid "Foo bar baz"\n' 'msgstr ""\n'
        )
        fn = tmpdir.join("messages.po")
        fn.write(po_file)
        # Verify the last line is untranslated
        assert po_file.endswith('msgstr ""\n')

        result = runner.invoke(cli, ("translate", "-p", "shouty", str(fn)))
        assert result.exit_code == 0
        last_line = fn.read().splitlines()[-1]
        # Verify the last line is now translated
        assert last_line == 'msgstr "FOO BAR BAZ"'

    def test_bad_pipeline(self, runner, tmpdir):
        po_file = build_po_string(
            "#: foo/foo.py:5\n" 'msgid "Foo bar baz"\n' 'msgstr ""\n'
        )
        fn = tmpdir.join("messages.po")
        fn.write(po_file)

        result = runner.invoke(cli, ("translate", "-p", "triangle", str(fn)))
        assert result.exit_code == 2
        last_line = result.output.splitlines()[-1]
        assert last_line == 'Error: pipeline "triangle" is not valid'

    def test_nonexistent_file(self, runner, tmpdir):
        fn = tmpdir.join("missingfile.po")
        result = runner.invoke(cli, ("translate", "-p", "shouty", str(fn)))
        assert result.exit_code == 2
        last_line = result.output.splitlines()[-1]
        assert last_line == "Error: File %s does not exist." % str(fn)

    def test_plurals(self, runner, tmpdir):
        po_file = build_po_string(
            "#: test_project/base/views.py:12 test_project/base/views.py:22\n"
            "#, python-format\n"
            'msgid "%(num)s apple"\n'
            'msgid_plural "%(num)s apples"\n'
            'msgstr[0] ""\n'
            'msgstr[1] ""\n'
        )
        fn = tmpdir.join("messages.po")
        fn.write(po_file)

        result = runner.invoke(cli, ("translate", "-p", "shouty", str(fn)))
        assert result.exit_code == 0
        pot_file = nix_header(fn.read())
        # FIXME: This has the wrong variables, too. We need to fix that, too.
        assert pot_file == dedent("""\
            #: test_project/base/views.py:12 test_project/base/views.py:22
            #, python-format
            msgid "%(num)s apple"
            msgid_plural "%(num)s apples"
            msgstr[0] "%(NUM)S APPLE"
            msgstr[1] "%(NUM)S APPLES"
            """)


class TestLint:
    def test_help(self, runner):
        result = runner.invoke(cli, ("lint", "--help"))
        assert result.exit_code == 0
        assert "Usage" in result.output.splitlines()[0]

    def test_basic_linting(self, runner, tmpdir):
        po_file = build_po_string(
            "#: foo/foo.py:5\n" 'msgid "Foo bar baz"\n' 'msgstr "Foo"\n'
        )
        fn = tmpdir.join("messages.po")
        fn.write(po_file)

        result = runner.invoke(cli, ("lint", str(fn)))
        assert result.exit_code == 0
        output_lines = result.output.splitlines()
        assert len(output_lines) == 1
        assert output_lines[0].startswith("dennis version")

    def test_linting_non_ascii_filename(self, runner, tmpdir):
        po_file = build_po_string(
            "#: foo/foo.py:5\n" 'msgid "Foo %(foo)s bar baz"\n' 'msgstr "Foo %(bar)s"\n'
        )
        fn = tmpdir.join("T\xc3\xa9l\xc3\xa9chargements", "messages.po")
        fn.write(po_file, ensure=True)

        result = runner.invoke(cli, ("lint", str(tmpdir)))
        assert result.exit_code == 1

    def test_linting_fail(self, runner, tmpdir):
        po_file = build_po_string(
            "#: foo/foo.py:5\n" 'msgid "Foo %(foo)s bar baz"\n' 'msgstr "Foo %(bar)s"\n'
        )
        fn = tmpdir.join("messages.po")
        fn.write(po_file)

        result = runner.invoke(cli, ("lint", str(fn)))
        assert result.exit_code == 1
        output_lines = result.output.splitlines()
        assert len(output_lines) == 16
        assert "E201: invalid variables: %(bar)s" in result.output

    def test_no_files_to_work_on(self, runner):
        result = runner.invoke(cli, ("lint", "foo"))
        assert result.exit_code == 2
        assert "nothing to work on." in result.output

    def test_nonexistent_file(self, runner, tmpdir):
        fn = tmpdir.join("missingfile.po")
        result = runner.invoke(cli, ("lint", str(fn)))
        assert result.exit_code == 2
        last_line = result.output.splitlines()[-1]
        assert last_line == 'Error: File "%s" does not exist.' % str(fn)

    def test_quiet(self, runner, tmpdir):
        po_file = build_po_string(
            "#: foo/foo.py:5\n" 'msgid "Foo bar baz"\n' 'msgstr "Foo"\n'
        )
        fn = tmpdir.join("messages.po")
        fn.write(po_file)

        result = runner.invoke(cli, ("lint", "--quiet", str(fn)))
        assert result.exit_code == 0
        assert result.output == ""

    def test_rules(self, runner, tmpdir):
        po_file = build_po_string(
            "#: foo/foo.py:5\n" 'msgid "Foo %(bar)s baz"\n' 'msgstr "FOO %(bar) BAZ"\n'
        )
        fn = tmpdir.join("messages.po")
        fn.write(po_file)

        result = runner.invoke(cli, ("lint", "--rules", "E101", str(fn)))
        assert result.exit_code == 1
        assert "E101: type missing: %(bar)" in result.output

    def test_template_rules(self, runner, tmpdir):
        po_file = build_po_string(
            "#: foo/foo.py:5\n" 'msgid "Foo %(o)s baz"\n' 'msgstr ""\n'
        )
        fn = tmpdir.join("messages.pot")
        fn.write(po_file)

        result = runner.invoke(cli, ("lint", "--rules", "W501", str(fn)))
        assert result.exit_code == 0
        assert 'W501: one character variable name "o"' in result.output

    def test_nonexisting_rule(self, runner, tmpdir):
        po_file = build_po_string(
            "#: foo/foo.py:5\n" 'msgid "Foo %(o)s baz"\n' 'msgstr ""\n'
        )
        fn = tmpdir.join("messages.pot")
        fn.write(po_file)

        result = runner.invoke(cli, ("lint", "--rules", "foo", str(fn)))
        assert result.exit_code == 2
        assert "Error: invalid rules: foo." in result.output

    def test_excludes(self, runner, tmpdir):
        po_file = build_po_string(
            "#: foo/foo.py:5\n" 'msgid "Foo %(o)s baz"\n' 'msgstr ""\n'
        )
        fn = tmpdir.join("messages.pot")
        fn.write(po_file)

        result = runner.invoke(cli, ("lint", "--excluderules", "W501", str(fn)))
        assert result.exit_code == 0
        # The rule that generates this error is excluded, so this error shouldn't show up.
        assert 'W501: one character variable name "o"' not in result.output

    def test_varformat_no_value(self, runner, tmpdir):
        po_file = build_po_string(
            "#: foo/foo.py:5\n" 'msgid "Foo %(o)s baz"\n' 'msgstr ""\n'
        )
        fn = tmpdir.join("messages.pot")
        fn.write(po_file)

        result = runner.invoke(cli, ("lint", "--varformat", "", str(fn)))

        assert result.exit_code == 0
        # We're not looking at any variables, so we should never have a variable related warning.
        assert 'W501: one character variable name "o"' not in result.output

    # FIXME: test --varformat with values

    # FIXME: test --reporter

    # FIXME: test --errorsonly



import pytest

from dennis.tools import VariableTokenizer
from dennis.translator import (
    AngleQuoteTransform,
    DubstepTransform,
    EmptyTransform,
    HTMLExtractorTransform,
    HahaTransform,
    PirateTransform,
    RedactedTransform,
    ReverseTransform,
    ShoutyTransform,
    Token,
    Translator,
    DoubleTransform,
    XXXTransform,
    ZombieTransform,
)


class TransformTestCase:
    vartok = VariableTokenizer(["python-format", "python-brace-format"])


class TestEmptyTransform(TransformTestCase):
    @pytest.mark.parametrize("text,expected", [("Hello", ""), ("OMG!", ""), ("", "")])
    def test_basic(self, text, expected):
        trans = EmptyTransform()
        output = trans.transform(self.vartok, [Token(text)])
        output = "".join([token.s for token in output])

        assert output == expected


class TestHahaTransform(TransformTestCase):
    @pytest.mark.parametrize(
        "text,expected",
        [
            ("Hello", "Haha\u2757 Hello"),
            ("OMG! Hello!", "Haha\u2757 OMG! Haha\u2757 Hello!"),
            ("", "Haha\u2757 "),
            ("Hi!\nHello!", "Haha\u2757 Hi!\nHaha\u2757 Hello!"),
        ],
    )
    def test_basic(self, text, expected):
        trans = HahaTransform()
        output = trans.transform(self.vartok, [Token(text)])
        output = "".join([token.s for token in output])

        assert output == expected


class TestPirateTransform(TransformTestCase):
    @pytest.mark.parametrize(
        "text,expected",
        [
            ("Hello", "'ello ahoy\u2757"),
            ("Hello %(username)s", "'ello %(username)s ahoy\u2757"),
            ("Hello %s", "'ello %s cap'n\u2757"),
            ("Hello {user}{name}", "'ello {user}{name} ahoy\u2757"),
            ("Products and Services", "Products and Sarrvices yo-ho-ho\u2757"),
            ("Get community support", "Get community supparrt yo-ho-ho\u2757"),
            (
                "Your input helps make Mozilla better",
                "Yerr input 'elps make Mozillar betterr prepare to be boarded\u2757",
            ),
            ("Super browsing", "Superr browsin' arrRRRrrr\u2757"),
        ],
    )
    def test_basic(self, text, expected):
        trans = PirateTransform()
        output = trans.transform(self.vartok, [Token(text)])
        output = "".join([token.s for token in output])

        assert output == expected


class TestDubstepTransform(TransformTestCase):
    @pytest.mark.parametrize(
        "text,expected",
        [
            ("Hello", "Hello t-t-t-t\u2757"),
            (
                "Hello %(username)s",
                "Hello t-t-t-t %(username)s BWWWWWWAAAAAAAAaaaaaa\u2757",
            ),
            ("Hello %s", "Hello t-t-t-t %s BWWWWAAAAAaaaa\u2757"),
            (
                "Hello {user}{name}",
                "Hello t-t-t-t {user}{name} BWWWWWWAAAAAAAAaaaaaa\u2757",
            ),
            ("Products and Services", "Products and Services BWWWWWAAAAAAAaaaaa\u2757"),
            (
                "Get community support",
                (
                    "Get t-t-t-t community BWWWWWAAAAAAAaaaaa support V\u221eP V\u221eP\u2757"
                ),
            ),
            (
                "Your input helps make Mozilla better",
                (
                    "Your input helps make BWWWWWAAAAAAAaaaaa Mozilla "
                    "....vvvVV better BWWWWWWAAAAAAAAaaaaaa\u2757"
                ),
            ),
            ("Super browsing", "Super t-t-t-t browsing BWWWWWAAAAAAAaaaaa\u2757"),
            (
                "Sentence 1. Sentence 2!",
                "Sentence 1 . t-t-t-t Sentence 2! BWWWWWWWWAAAAAAAAAAaaaaaaaa\u2757",
            ),
        ],
    )
    def test_basic(self, text, expected):
        trans = DubstepTransform()
        output = trans.transform(self.vartok, [Token(text)])
        output = "".join([token.s for token in output])

        assert output == expected


class TestZombieTransform(TransformTestCase):
    @pytest.mark.parametrize(
        "text,expected",
        [
            ("", ""),
            ("Hello", "HHAMNMNHR RARRR!!!"),
            ("Hi\nHello", "HAR\nHHAMNMNHR BR-R-R-RAINS!"),
            ("Hi   \nHello!\nmultiline", "HAR   \nHHAMNMNHR\u2757\nmNMMNHGARMNARnHA"),
            ("Hello %(username)s", "HHAMNMNHR %(username)s"),
            ("Hello %s", "HHAMNMNHR %s GRRRRRrrRR!!"),
            # FIXME: This is a problem where variables from the variable tokenizer should
            # be treated as immutable tokens and right now they aren't. Issue #67.
            ("Hello {user}{name}", "HHAMNMNHR {user}{namHA}"),
            ("Products and Services", "PMZHRGBNMZZHGRZ anGB SHAMZBBARZZHARZ"),
            ("Get community support", "GHAHG ZZHRmmNMnARHGRA RZNMBZBZHRMZHG"),
            (
                "Your input helps make Mozilla better",
                (
                    "YHRNMMZ ARnBZNMHG hHAMNBZRZ maBGHA MHRzARMNMNa bHAHGHGHAMZ "
                    "BR-R-R-RAINS!"
                ),
            ),
            ("Super browsing", "SNMBZHAMZ bMZHRZMRZARng"),
        ],
    )
    def test_basic(self, text, expected):
        trans = ZombieTransform()
        output = trans.transform(self.vartok, [Token(text)])
        output = "".join([token.s for token in output])

        assert output == expected


class TestHTMLExtractor:
    vartok = VariableTokenizer(["python-format", "python-brace-format"])

    def test_basic(self):
        trans = HTMLExtractorTransform()
        output = trans.transform(self.vartok, [Token("")])
        assert output == [Token("", "text", True)]

        output = trans.transform(self.vartok, [Token("<b>hi</b>")])
        assert output == [
            Token("<b>", "html", False),
            Token("hi", "text", True),
            Token("</b>", "html", False),
        ]

    def test_alt_title_placeholder(self):
        trans = HTMLExtractorTransform()
        output = trans.transform(self.vartok, [Token('<img alt="foo">')])
        assert output == [
            Token('<img alt="', "html", False),
            Token("foo", "text", True),
            Token('">', "html", False),
        ]

        output = trans.transform(self.vartok, [Token('<img title="foo">')])
        assert output == [
            Token('<img title="', "html", False),
            Token("foo", "text", True),
            Token('">', "html", False),
        ]

        output = trans.transform(self.vartok, [Token('<input placeholder="foo">')])
        assert output == [
            Token('<input placeholder="', "html", False),
            Token("foo", "text", True),
            Token('">', "html", False),
        ]

    def test_script_style(self):
        trans = HTMLExtractorTransform()
        output = trans.transform(
            self.vartok, [Token("<style>TR {white-space: nowrap;}</style>")]
        )
        assert output == [
            Token("<style>", "html", False),
            Token("TR {white-space: nowrap;}", "style", False),
            Token("</style>", "html", False),
        ]

        output = trans.transform(
            self.vartok, [Token('<script>console.log("foo");</script>')]
        )
        assert output == [
            Token("<script>", "html", False),
            Token('console.log("foo");', "script", False),
            Token("</script>", "html", False),
        ]

    def test_whitespace_collapse(self):
        def _trans(text):
            return HTMLExtractorTransform().transform(self.vartok, [Token(text)])

        assert _trans("<html><body>hello!</body></html>") == _trans(
            "<html><body>    hello!    </body></html>"
        )


class TestXXXTransform(TransformTestCase):
    @pytest.mark.parametrize(
        "text,expected",
        [
            ("Hello", "xxxHelloxxx"),
            ("OMG!", "xxxOMG!xxx"),
            ("Line\nWith\nCRs", "xxxLinexxx\nxxxWithxxx\nxxxCRsxxx"),
            ("Line.  \nWith!\nCRs", "xxxLine.xxx  \nxxxWith!xxx\nxxxCRsxxx"),
        ],
    )
    def test_basic(self, text, expected):
        trans = XXXTransform()
        output = trans.transform(self.vartok, [Token(text)])
        output = "".join([token.s for token in output])

        assert output == expected


class TestDoubleTransform(TransformTestCase):
    @pytest.mark.parametrize(
        "text,expected",
        [
            ("Hello", "Heelloo"),
            ("OMG!", "OOMG!"),
            ("Line\nWith\nCRs", "Liinee\nWiith\nCRs"),
        ],
    )
    def test_basic(self, text, expected):
        trans = DoubleTransform()
        output = trans.transform(self.vartok, [Token(text)])
        output = "".join([token.s for token in output])

        assert output == expected


class TestAngleQuoteTransform(TransformTestCase):
    @pytest.mark.parametrize(
        "text,expected",
        [
            ("Hello", "\xabHello\xbb"),
            ("OMG!", "\xabOMG!\xbb"),
            ("Line\nWith\nCRs", "\xabLine\nWith\nCRs\xbb"),
            ("Line.  \nWith!\nCRs", "\xabLine.  \nWith!\nCRs\xbb"),
        ],
    )
    def test_basic(self, text, expected):
        trans = AngleQuoteTransform()
        output = trans.transform(self.vartok, [Token(text)])
        output = "".join([token.s for token in output])

        assert output == expected


class TestShoutyTransform(TransformTestCase):
    @pytest.mark.parametrize(
        "text,expected",
        [("Hello", "HELLO"), ("OMG!", "OMG!"), ("<b>Hello.</b>", "<B>HELLO.</B>")],
    )
    def test_basic(self, text, expected):
        trans = ShoutyTransform()
        output = trans.transform(self.vartok, [Token(text)])
        output = "".join([token.s for token in output])

        assert output == expected


class TestReverseTransform(TransformTestCase):
    @pytest.mark.parametrize(
        "text,expected",
        [
            ("Hello", "olleH"),
            ("OMG!", "!GMO"),
            ("Hello. This is a test.", ".tset a si sihT .olleH"),
        ],
    )
    def test_basic(self, text, expected):
        trans = ReverseTransform()
        output = trans.transform(self.vartok, [Token(text)])
        output = "".join([token.s for token in output])

        assert output == expected


class TestRedactedTransform(TransformTestCase):
    @pytest.mark.parametrize("text,expected", [("Hello", "Xxxxx"), ("OMG!", "XXX!")])
    def test_basic(self, text, expected):
        trans = RedactedTransform()
        output = trans.transform(self.vartok, [Token(text)])
        output = "".join([token.s for token in output])

        assert output == expected


class TestTranslator:
    def test_pirate_translate(self):
        trans = Translator(["python-format", "python-brace-format"], ["pirate"])
        assert trans.translate_string("hello") == "'ello ahoy\u2757"

        # Note, this doesn't work right because it's not accounting
        # for html. But it's correct for this test.
        assert (
            trans.translate_string("<b>hello</b>")
            == "<b>'ello</b> prepare to be boarded\u2757"
        )

    def test_html_pirate_translate(self):
        trans = Translator(["python-format", "python-brace-format"], ["html", "pirate"])
        assert trans.translate_string("<b>hello</b>") == "<b>'ello ahoy\u2757</b>"

    def test_shouty_html_pirate_translate(self):
        trans = Translator(
            ["python-format", "python-brace-format"], ["shouty", "html", "pirate"]
        )
        assert trans.translate_string("<b>hello.</b>\n") == "<b>HELLO aye\u2757.</b>"



#######################################################################
# This file is part of dennis.
#
# Copyright (C) 2013-2022 Will Kahn-Greene
# Licensed under the Simplified BSD License. See LICENSE for full
# license.
#######################################################################

from setuptools import setup, find_packages
import re
import os

READMEFILE = "README.rst"
VERSIONFILE = os.path.join("dennis", "__init__.py")
VSRE = r"""^__version__ = ['"]([^'"]*)['"]"""


def get_version():
    version_file = open(VERSIONFILE, "rt").read()
    return re.search(VSRE, version_file, re.M).group(1)


INSTALL_REQUIRES = [
    "polib>=1.0.8",
    "click>=6",
]

setup(
    name="dennis",
    version=get_version(),
    description=(
        "Utilities for working with gettext PO and POT files to ease development "
        "and improve localization quality"
    ),
    long_description=open(READMEFILE).read(),
    license="Simplified BSD License",
    maintainer="Mozilla",
    author="Will Kahn-Greene",
    keywords="l10n localization gettext PO POT lint translate development",
    url="https://github.com/mozilla/dennis",
    zip_safe=True,
    packages=find_packages(),
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    entry_points="""
        [console_scripts]
        dennis-cmd=dennis.cmdline:click_run
    """,
    python_requires=">=3.10",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Localization",
    ],
)



import re
import uuid
from collections import namedtuple
from itertools import zip_longest

from dennis.tools import (
    VariableTokenizer,
    all_subclasses,
    parse_dennis_note,
    parse_pofile,
)

IdString = namedtuple("IdString", ("msgid_fields", "msgid_strings"))
TranslatedString = namedtuple(
    "TranslatedString",
    ("msgid_fields", "msgid_strings", "msgstr_field", "msgstr_string"),
)


WARNING = "warn"
ERROR = "err"


class HTMLParseError(Exception):
    pass


class LintMessage:
    def __init__(self, kind, line, col, code, msg, poentry):
        self.kind = kind
        self.line = line
        self.col = col
        self.code = code
        self.msg = msg
        self.poentry = poentry

    def __repr__(self):
        return "<LintedMessage {} {}:{} {} {}>".format(
            self.kind, self.line, self.col, self.code, self.msg
        )


class LintedEntry:
    def __init__(self, poentry):
        self.poentry = poentry
        self.msgid = poentry.msgid

    @property
    def str(self):
        poentry = self.poentry
        if poentry.msgid_plural:
            msgid_fields = ("msgid", "msgid_plural")
            msgid_strings = (poentry.msgid, poentry.msgid_plural)
        else:
            msgid_fields = ("msgid",)
            msgid_strings = (poentry.msgid,)

        return IdString(msgid_fields, msgid_strings)

    @property
    def strs(self):
        poentry = self.poentry
        strs = []

        if not poentry.msgid_plural:
            strs.append(
                TranslatedString(["msgid"], [poentry.msgid], "msgstr", poentry.msgstr)
            )

        else:
            msgid_fields = ("msgid", "msgid_plural")
            msgid_strings = (poentry.msgid, poentry.msgid_plural)

            for key in sorted(poentry.msgstr_plural.keys()):
                strs.append(
                    TranslatedString(
                        msgid_fields,
                        msgid_strings,
                        "msgstr[{}]".format(key),
                        poentry.msgstr_plural[key],
                    )
                )

        # List of TranslatedStrings
        return strs


class LintRule:
    num = ""
    name = ""
    desc = ""

    def lint(self, vartok, linted_entry):
        """Takes a linted entry and generates LintMessages

        :arg vartok: the variable tokenizer for extracting variables
        :arg linted_entry: the LintedEntry to work on

        :returns: list of LintMeessages or empty list

        """
        raise NotImplementedError


class MalformedNoTypeLintRule(LintRule):
    num = "E101"
    name = "notype"
    desc = "%(count) with no type at the end"

    def lint(self, vartok, linted_entry):
        msgs = []

        # This only applies if one of the variable tokenizers is python-format.
        if not vartok.contains("python-format"):
            return msgs

        malformed_re = re.compile(
            r"(?:"
            r"%"  # %
            r"[\(][^\)\s]+[\)]"  # things in parens or not
            r"(?:(?=[^diouxefGgcrs])|$)"  # end of string or something that's not a format char
            r")"
        )

        for trstr in linted_entry.strs:
            if not trstr.msgstr_string:
                continue

            malformed = malformed_re.findall(trstr.msgstr_string)
            if not malformed:
                continue

            malformed = [item.strip() for item in malformed]
            msgs.append(
                LintMessage(
                    ERROR,
                    linted_entry.poentry.linenum,
                    0,
                    self.num,
                    "type missing: {}".format(", ".join(malformed)),
                    linted_entry.poentry,
                )
            )

        return msgs


class MalformedMissingRightBraceLintRule(LintRule):
    num = "E102"
    name = "missingrightbrace"
    desc = "{foo with missing }"

    def lint(self, vartok, linted_entry):
        msgs = []

        # This only applies if one of the variable tokenizers is python-brace-format.
        if not vartok.contains("python-brace-format"):
            return []

        malformed_re = re.compile(r"(?:\{[^\}]+(?:\{|$))")
        double_open = str(uuid.uuid4())
        double_close = str(uuid.uuid4())

        for trstr in linted_entry.strs:
            if not trstr.msgstr_string:
                continue

            malformed = malformed_re.findall(
                trstr.msgstr_string.replace("{{", double_open).replace(
                    "}}", double_close
                )
            )
            if not malformed:
                continue

            malformed = [
                item.strip().replace(double_open, "{{").replace(double_close, "}}")
                for item in malformed
            ]
            msgs.append(
                LintMessage(
                    ERROR,
                    linted_entry.poentry.linenum,
                    0,
                    self.num,
                    "missing right curly-brace: {}".format(", ".join(malformed)),
                    linted_entry.poentry,
                )
            )

        return msgs


class MalformedMissingLeftBraceLintRule(LintRule):
    num = "E103"
    name = "missingleftbrace"
    desc = "foo} with missing {"

    def lint(self, vartok, linted_entry):
        msgs = []

        # This only applies if one of the variable tokenizers is python-brace-format.
        if not vartok.contains("python-brace-format"):
            return []

        malformed_re = re.compile(r"(?:(?:^|\})[^\{]*\})")
        double_open = str(uuid.uuid4())
        double_close = str(uuid.uuid4())

        for trstr in linted_entry.strs:
            if not trstr.msgstr_string:
                continue

            malformed = malformed_re.findall(
                trstr.msgstr_string.replace("{{", double_open).replace(
                    "}}", double_close
                )
            )
            if not malformed:
                continue

            malformed = [
                item.strip().replace(double_open, "{{").replace(double_close, "}}")
                for item in malformed
            ]
            msgs.append(
                LintMessage(
                    ERROR,
                    linted_entry.poentry.linenum,
                    0,
                    self.num,
                    "missing left curly-brace: {}".format(", ".join(malformed)),
                    linted_entry.poentry,
                )
            )
        return msgs


class BadFormatLintRule(LintRule):
    num = "E104"
    name = "badformat"
    desc = "% followed by a bad format character"

    def lint(self, vartok, linted_entry):
        msgs = []

        # This only applies if one of the variable tokenizers is python-format.
        if not vartok.contains("python-format"):
            return []

        splitter = re.compile(r"(\%(?:.|$))")

        for trstr in linted_entry.strs:
            if not trstr.msgstr_string:
                continue

            msgid_tokens = list(vartok.extract_tokens(" ".join(trstr.msgid_strings)))
            if not msgid_tokens:
                continue
            # FIXME: Pretty sure we can just check the first token because they'll
            # all have to be the same kind of thing.
            first_token = msgid_tokens[0]
            if not (len(first_token) == 2 and first_token[0] == "%"):
                continue

            for match in splitter.findall(trstr.msgstr_string):
                if len(match) == 1 or match[1] not in "(diouxefGgcrs%":
                    msgs.append(
                        LintMessage(
                            ERROR,
                            linted_entry.poentry.linenum,
                            0,
                            self.num,
                            "bad format character: {}".format(match),
                            linted_entry.poentry,
                        )
                    )
        return msgs


class MissingVarsLintRule(LintRule):
    num = "W202"
    num_error = "E202"
    name = "missingvars"
    desc = "Checks for variables in msgid, but missing in msgstr"

    def lint(self, vartok, linted_entry):
        msgs = []

        # If there are no variable formats, skip this rule.
        if not vartok.formats:
            return []

        for trstr in linted_entry.strs:
            if not trstr.msgstr_string:
                continue

            if len(trstr.msgid_strings) > 1:
                # If the msgstr uses a variable that isn't in any of
                # the msgid strings, then that's an error (and handled
                # by a different LintRule). Anything else we should
                # ignore.
                continue

            msgid_tokens = vartok.extract_tokens(" ".join(trstr.msgid_strings))
            msgstr_tokens = vartok.extract_tokens(trstr.msgstr_string)

            missing = msgid_tokens.difference(msgstr_tokens)

            if missing:
                if vartok.contains("python-format") and len(
                    [var for var in missing if len(var) == 2 and var[0] == "%"]
                ):
                    # If we're looking at python-format variables and one of these is a
                    # python-format variable like %s, then this is an error--not a warning.
                    msgs.append(
                        LintMessage(
                            ERROR,
                            linted_entry.poentry.linenum,
                            0,
                            self.num_error,
                            "missing variables: {}".format(", ".join(sorted(missing))),
                            linted_entry.poentry,
                        )
                    )
                else:
                    # If this is not python-format variables, then this is just a warning.
                    msgs.append(
                        LintMessage(
                            WARNING,
                            linted_entry.poentry.linenum,
                            0,
                            self.num,
                            "missing variables: {}".format(", ".join(sorted(missing))),
                            linted_entry.poentry,
                        )
                    )
        return msgs


class BlankLintRule(LintRule):
    num = "W301"
    name = "blank"
    desc = "Translated string is only whitespace"

    def lint(self, vartok, linted_entry):
        msgs = []

        for trstr in linted_entry.strs:
            if not trstr.msgstr_string:
                continue

            if trstr.msgstr_string.isspace():
                msgs.append(
                    LintMessage(
                        WARNING,
                        linted_entry.poentry.linenum,
                        0,
                        self.num,
                        "translated string is solely whitespace",
                        linted_entry.poentry,
                    )
                )

        return msgs


class UnchangedLintRule(LintRule):
    num = "W302"
    name = "unchanged"
    desc = "Makes sure string is actually translated"

    def lint(self, vartok, linted_entry):
        msgs = []

        for trstr in linted_entry.strs:
            if not trstr.msgstr_string:
                continue

            if trstr.msgstr_string in trstr.msgid_strings:
                msgs.append(
                    LintMessage(
                        WARNING,
                        linted_entry.poentry.linenum,
                        0,
                        self.num,
                        "translated string is same as source string",
                        linted_entry.poentry,
                    )
                )

        return msgs


class MismatchedHTMLLintRule(LintRule):
    num = "W303"
    num_error = "W304"
    name = "html"
    desc = "Checks for matching html between source and translated strings"

    def lint(self, vartok, linted_entry):
        msgs = []

        from dennis.translator import HTMLExtractorTransform, Token

        html = HTMLExtractorTransform()

        def equiv(left, right):
            return left == right

        def tokenize(text):
            """Tokenizes the text using the HTMLExtractorTransform

            :raises HTMLParseError: If it's invalid HTML.

            """
            tokens = [
                token
                for token in html.transform(vartok, [Token(text)])
                if token.type == "html" and not token.s.startswith("&")
            ]
            return sorted(tokens, key=lambda token: token.s)

        for trstr in linted_entry.strs:
            if not trstr.msgstr_string:
                continue

            try:
                msgid_parts = tokenize(trstr.msgid_strings[0])
            except HTMLParseError as exc:
                errmsg = "invalid html: msgid has invalid html {}".format(exc)
                msgs.append(
                    LintMessage(
                        WARNING,
                        linted_entry.poentry.linenum,
                        0,
                        self.num_error,
                        errmsg,
                        linted_entry.poentry,
                    )
                )
                return msgs

            if len(trstr.msgid_strings) > 1:
                # If this is a plural, then we check to see if the two
                # msgid strings match each other. If not, then we move
                # on because I have no idea what to do in this case.
                try:
                    msgid_plural_parts = tokenize(trstr.msgid_strings[1])
                except HTMLParseError as exc:
                    errmsg = "invalid html: msgid_plural has invalid html {}".format(
                        exc
                    )

                    msgs.append(
                        LintMessage(
                            WARNING,
                            linted_entry.poentry.linenum,
                            0,
                            self.num_error,
                            errmsg,
                            linted_entry.poentry,
                        )
                    )
                    return msgs

                zipped_parts = zip_longest(
                    msgid_parts, msgid_plural_parts, fillvalue=None
                )

                for left, right in zipped_parts:
                    if not left or not right or not equiv(left, right):
                        return []

            try:
                msgstr_parts = tokenize(trstr.msgstr_string)
            except HTMLParseError as exc:
                errmsg = "invalid html: msgstr has invalid html {}".format(exc)
                msgs.append(
                    LintMessage(
                        WARNING,
                        linted_entry.poentry.linenum,
                        0,
                        self.num_error,
                        errmsg,
                        linted_entry.poentry,
                    )
                )
                return msgs

            for left, right in zip_longest(msgid_parts, msgstr_parts, fillvalue=None):
                if not left or not right or not equiv(left, right):
                    msgs.append(
                        LintMessage(
                            WARNING,
                            linted_entry.poentry.linenum,
                            0,
                            self.num,
                            'different html: "{}" vs. "{}"'.format(
                                left.s if left else "", right.s if right else ""
                            ),
                            linted_entry.poentry,
                        )
                    )
                    break
        return msgs


class InvalidVarsLintRule(LintRule):
    num = "E201"
    name = "invalidvars"
    desc = "Checks for variables not in msgid, but in msgstr"

    def lint(self, vartok, linted_entry):
        # If there are no variable formats, skip this rule.
        if not vartok.formats:
            return []

        msgs = []
        for trstr in linted_entry.strs:
            if not trstr.msgstr_string:
                continue

            msgid_tokens = vartok.extract_tokens(" ".join(trstr.msgid_strings))
            # If this is python-format or python-brace-format and there are
            # no tokens in the msgid, then "no tokens, no problem".
            if not msgid_tokens:
                continue

            msgstr_tokens = vartok.extract_tokens(trstr.msgstr_string)

            invalid = msgstr_tokens.difference(msgid_tokens)

            if invalid:
                msgs.append(
                    LintMessage(
                        ERROR,
                        linted_entry.poentry.linenum,
                        0,
                        self.num,
                        "invalid variables: {}".format(", ".join(sorted(invalid))),
                        linted_entry.poentry,
                    )
                )
        return msgs


def get_lint_rules(with_names=False):
    lint_rules = {}

    for cls in all_subclasses(LintRule):
        if cls.num:
            lint_rules[cls.num] = cls
            if with_names and cls.name:
                lint_rules[cls.name] = cls

    return lint_rules


def convert_rules(rules_spec):
    lint_rules = get_lint_rules(with_names=True)
    rules = [lint_rules[rule]() for rule in rules_spec if rule in lint_rules]
    return rules


class Linter:
    def __init__(self, vars_, rules_spec):
        self.vartok = VariableTokenizer(vars_)
        self.rules_spec = rules_spec
        self.rules = convert_rules(self.rules_spec)

    def lint_poentry(self, poentry):
        linted_entry = LintedEntry(poentry)

        skip = parse_dennis_note(poentry.comment)

        msgs = []

        # Check the comment to see if what we should ignore.
        for lint_rule in self.rules:
            if skip == "*" or lint_rule.num in skip:
                continue

            msgs.extend(lint_rule.lint(self.vartok, linted_entry))

        return msgs

    def verify_file(self, filename_or_string):
        """Verifies strings in file.

        :arg filename_or_string: filename to verify or the contents of
            a pofile as a string

        :returns: list of LintMessage objects

        :raises IOError: if the file is not a valid .po file or
            doesn't exist
        """
        po = parse_pofile(filename_or_string)
        msgs = []
        for entry in po.translated_entries():
            msgs.extend(self.lint_poentry(entry))

        return msgs
